from abc import ABC, abstractmethod


class Agent(ABC):
    @abstractmethod
    def get_action(self, obs): ...


from .memory import MemoryAgent
from .running import SyncRunningAgent
from .actor_critic import ActorCriticAgent, DEFAULTS



import gin
import tensorflow.compat.v1 as tf
from tensorflow.compat.v1.keras.layers import Input, Concatenate, Dense
from reaver.models.base.layers import Squeeze, Variable, RunningStatsNorm
from reaver.envs.base import Spec


@gin.configurable
def build_mlp(
        obs_spec: Spec,
        act_spec: Spec,
        layer_sizes=(64, 64),
        activation='relu',
        initializer='glorot_uniform',
        value_separate=False,
        obs_shift=False,
        obs_scale=False) -> tf.keras.Model:
    """
    Factory method for a simple fully connected neural network model used in e.g. MuJuCo environment

    If value separate is set to true then a separate path is added for value fn, otherwise branches out of last layer
    If obs shift is set to true then observations are normalized to mean zero with running mean estimate
    If obs scale is set to true then observations are standardized to std.dev one with running std.dev estimate
    """
    inputs = inputs_ = [Input(s.shape, name="input_" + s.name) for s in obs_spec]
    if obs_shift or obs_scale:
        inputs_ = [RunningStatsNorm(obs_shift, obs_scale, name="norm_" + s.name)(x) for s, x in zip(obs_spec, inputs_)]
    inputs_concat = Concatenate()(inputs_) if len(inputs_) > 1 else inputs_[0]

    x = build_fc(inputs_concat, layer_sizes, activation, initializer)
    outputs = [build_logits(space, x, initializer) for space in act_spec]

    if value_separate:
        x = build_fc(inputs_concat, layer_sizes, activation, initializer, 'value_')

    value = Dense(1, name="value_out", kernel_initializer=initializer)(x)
    value = Squeeze(axis=-1)(value)
    outputs.append(value)

    return tf.keras.Model(inputs=inputs, outputs=outputs)


def build_logits(space, prev_layer, initializer):
    logits = Dense(space.size(), kernel_initializer=initializer, name="logits_" + space.name)(prev_layer)
    if space.is_continuous():
        logits = Variable(name="logstd")(logits)
    return logits


def build_fc(input_layer, layer_sizes, activation, initializer, prefix=''):
    x = input_layer
    for i, size in enumerate(layer_sizes):
        x = Dense(size, activation=activation, kernel_initializer=initializer, name='%sfc%02d' % (prefix, i+1))(x)
    return x



import gin.tf
import tensorflow.compat.v1 as tf

from reaver.envs.base import Spec
from reaver.utils import StreamLogger
from reaver.utils.tensorflow import SessionManager
from reaver.utils.typing import ModelBuilder, PolicyType
from reaver.agents.base import SyncRunningAgent, ActorCriticAgent, DEFAULTS


@gin.configurable('A2CAgent')
class AdvantageActorCriticAgent(SyncRunningAgent, ActorCriticAgent):
    """
    A2C: a synchronous version of Asynchronous Advantage Actor Critic (A3C)
    See article for more details: https://arxiv.org/abs/1602.01783
    """
    def __init__(
        self,
        obs_spec: Spec,
        act_spec: Spec,
        model_fn: ModelBuilder=None,
        policy_cls: PolicyType=None,
        sess_mgr: SessionManager=None,
        optimizer: tf.train.Optimizer=None,
        n_envs=4,
        value_coef=DEFAULTS['value_coef'],
        entropy_coef=DEFAULTS['entropy_coef'],
        traj_len=DEFAULTS['traj_len'],
        batch_sz=DEFAULTS['batch_sz'],
        discount=DEFAULTS['discount'],
        gae_lambda=DEFAULTS['gae_lambda'],
        clip_rewards=DEFAULTS['clip_rewards'],
        clip_grads_norm=DEFAULTS['clip_grads_norm'],
        normalize_returns=DEFAULTS['normalize_returns'],
        normalize_advantages=DEFAULTS['normalize_advantages'],
    ):
        kwargs = {k: v for k, v in locals().items() if k in DEFAULTS and DEFAULTS[k] != v}

        SyncRunningAgent.__init__(self, n_envs)
        ActorCriticAgent.__init__(self, obs_spec, act_spec, sess_mgr=sess_mgr, **kwargs)
        self.logger = StreamLogger(n_envs=n_envs, log_freq=10, sess_mgr=self.sess_mgr)

    def loss_fn(self):
        adv = tf.placeholder(tf.float32, [None], name="advantages")
        returns = tf.placeholder(tf.float32, [None], name="returns")

        policy_loss = -tf.reduce_mean(self.policy.logli * adv)
        value_loss = tf.reduce_mean((self.value - returns)**2) * self.value_coef
        entropy_loss = tf.reduce_mean(self.policy.entropy) * self.entropy_coef
        # we want to reduce policy and value errors, and maximize entropy
        # but since optimizer is minimizing the signs are opposite
        full_loss = policy_loss + value_loss - entropy_loss

        return full_loss, [policy_loss, value_loss, entropy_loss], [adv, returns]



import gin
from tensorflow.compat.v1.keras import Model
from tensorflow.compat.v1.keras.layers import Input, Concatenate, Dense, Conv2D, Flatten
from reaver.models.base.layers import Squeeze, Rescale, Transpose, RunningStatsNorm


@gin.configurable
def build_cnn_nature(obs_spec, act_spec, data_format='channels_first', value_separate=False, obs_shift=False, obs_scale=False):
    conv_cfg = dict(padding='same', data_format=data_format, activation='relu')
    conv_spec = [(32, 8, 4), (64, 4, 2), (64, 3, 1)]

    inputs = [Input(s.shape, name="input_" + s.name) for s in obs_spec]
    inputs_concat = Concatenate()(inputs) if len(inputs) > 1 else inputs[0]

    # expected NxCxHxW, but got NxHxWxC
    if data_format == 'channels_first' and inputs_concat.shape[1] > 3:
        inputs_concat = Transpose([0, 3, 1, 2])(inputs_concat)

    inputs_scaled = Rescale(1./255)(inputs_concat)
    if obs_shift or obs_scale:
        inputs_scaled = RunningStatsNorm(obs_shift, obs_scale)(inputs_scaled)

    x = build_cnn(inputs_scaled, conv_spec, conv_cfg, dense=512, prefix='policy_')
    outputs = [Dense(s.size(), name="logits_" + s.name)(x) for s in act_spec]

    if value_separate:
        x = build_cnn(inputs_scaled, conv_spec, conv_cfg, dense=512, prefix='value_')

    value = Dense(1, name="value_out")(x)
    value = Squeeze(axis=-1)(value)
    outputs.append(value)

    return Model(inputs=inputs, outputs=outputs)


def build_cnn(input_layer, layers, conv_cfg, dense=None, prefix=''):
    x = input_layer
    for i, (n_filters, kernel_size, stride) in enumerate(layers):
        x = Conv2D(n_filters, kernel_size, stride, name='%sconv%02d' % (prefix, i+1), **conv_cfg)(x)

    if dense:
        x = Flatten()(x)
        x = Dense(dense)(x)

    return x



import gin.tf
import numpy as np
import tensorflow.compat.v1 as tf

from reaver.envs.base import Spec
from reaver.utils import StreamLogger
from reaver.utils.tensorflow import SessionManager
from reaver.utils.typing import ModelBuilder, PolicyType
from reaver.agents.base import SyncRunningAgent, ActorCriticAgent, DEFAULTS


@gin.configurable('PPOAgent')
class ProximalPolicyOptimizationAgent(SyncRunningAgent, ActorCriticAgent):
    """
    PPO: clipped version of the Proximal Policy Optimization algorithm

    Here "clipped" refers to how trusted policy region is enforced.
    While orig. PPO relied on KL divergence, this clips the pi / pi_old ratio.

    See article for more details: https://arxiv.org/abs/1707.06347

    PPO specific parameters:

    :param n_epochs: number of times optimizer goes through full batch_sz*traj_len set
    :param minibatch_sz: size of the randomly sampled minibatch
    :param clip_ratio: max interval for pi / pi_old: [1-clip_ratio, 1+clip_ratio]
    :param clip_value: max interval for new value error: [old_value-clip_value, old_value+clip_value]
    """
    def __init__(
        self,
        obs_spec: Spec,
        act_spec: Spec,
        model_fn: ModelBuilder=None,
        policy_cls: PolicyType=None,
        sess_mgr: SessionManager=None,
        optimizer: tf.train.Optimizer=None,
        n_envs=4,
        n_epochs=3,
        minibatch_sz=128,
        clip_ratio=0.2,
        clip_value=0.5,
        value_coef=DEFAULTS['value_coef'],
        entropy_coef=DEFAULTS['entropy_coef'],
        traj_len=DEFAULTS['traj_len'],
        batch_sz=DEFAULTS['batch_sz'],
        discount=DEFAULTS['discount'],
        gae_lambda=DEFAULTS['gae_lambda'],
        clip_rewards=DEFAULTS['clip_rewards'],
        clip_grads_norm=DEFAULTS['clip_grads_norm'],
        normalize_returns=DEFAULTS['normalize_returns'],
        normalize_advantages=DEFAULTS['normalize_advantages'],
    ):
        kwargs = {k: v for k, v in locals().items() if k in DEFAULTS and DEFAULTS[k] != v}

        self.n_epochs = n_epochs
        self.minibatch_sz = minibatch_sz
        self.clip_ratio = clip_ratio
        self.clip_value = clip_value

        SyncRunningAgent.__init__(self, n_envs)
        ActorCriticAgent.__init__(self, obs_spec, act_spec, sess_mgr=sess_mgr, **kwargs)
        self.logger = StreamLogger(n_envs=n_envs, log_freq=10, sess_mgr=self.sess_mgr)

        self.start_step = self.start_step // self.n_epochs

    def minimize(self, advantages, returns):
        inputs = [a.reshape(-1, *a.shape[2:]) for a in self.obs + self.acts]
        tf_inputs = self.model.inputs + self.policy.inputs
        logli_old = self.sess_mgr.run(self.policy.logli, tf_inputs, inputs)

        inputs += [advantages.flatten(), returns.flatten(), logli_old, self.values.flatten()]
        tf_inputs += self.loss_inputs

        # TODO: rewrite this with persistent tensors to load data only once into the graph
        loss_terms = grads_norm = None
        n_samples = self.traj_len * self.batch_sz
        indices = np.arange(n_samples)
        for _ in range(self.n_epochs):
            np.random.shuffle(indices)
            for i in range(n_samples // self.minibatch_sz):
                idxs, idxe = i*self.minibatch_sz, (i+1)*self.minibatch_sz
                minibatch = [inpt[indices[idxs:idxe]] for inpt in inputs]
                loss_terms, grads_norm, *_ = self.sess_mgr.run(self.minimize_ops, tf_inputs, minibatch)

        return loss_terms, grads_norm

    def loss_fn(self):
        adv = tf.placeholder(tf.float32, [None], name="advantages")
        returns = tf.placeholder(tf.float32, [None], name="returns")
        logli_old = tf.placeholder(tf.float32, [None], name="logli_old")
        value_old = tf.placeholder(tf.float32, [None], name="value_old")

        ratio = tf.exp(self.policy.logli - logli_old)
        clipped_ratio = tf.clip_by_value(ratio, 1-self.clip_ratio, 1+self.clip_ratio)

        value_err = (self.value - returns)**2
        if self.clip_value > 0.0:
            clipped_value = tf.clip_by_value(self.value, value_old-self.clip_value, value_old+self.clip_value)
            clipped_value_err = (clipped_value - returns)**2
            value_err = tf.maximum(value_err, clipped_value_err)

        policy_loss = -tf.reduce_mean(tf.minimum(adv * ratio, adv * clipped_ratio))
        value_loss = tf.reduce_mean(value_err) * self.value_coef
        entropy_loss = tf.reduce_mean(self.policy.entropy) * self.entropy_coef
        # we want to reduce policy and value errors, and maximize entropy
        # but since optimizer is minimizing the signs are opposite
        full_loss = policy_loss + value_loss - entropy_loss

        return full_loss, [policy_loss, value_loss, entropy_loss], [adv, returns, logli_old, value_old]



import numpy as np
from .running import RunningAgent
from reaver.envs.base import Spec


class MemoryAgent(RunningAgent):
    """
    Handles experience memory storage - up to (B, T, ?),
    where B is batch size, T is trajectory length
    and ? is either 1 (None) for rewards/dones or act/obs shapes
    """
    def __init__(self, obs_spec: Spec, act_spec: Spec, traj_len, batch_sz):
        RunningAgent.__init__(self)

        self.traj_len = traj_len
        self.batch_sz = batch_sz
        self.shape = (traj_len, batch_sz)
        self.batch_ptr = 0
        self.n_batches = 0

        self.dones = np.empty(self.shape, dtype=np.bool)
        self.values = np.empty(self.shape, dtype=np.float32)
        self.rewards = np.empty(self.shape, dtype=np.float32)
        self.acts = [np.empty(self.shape + s.shape, dtype=s.dtype) for s in act_spec.spaces]
        self.obs = [np.empty(self.shape + s.shape, dtype=s.dtype) for s in obs_spec.spaces]
        self.last_obs = [np.empty((self.batch_sz, ) + s.shape, dtype=s.dtype) for s in obs_spec.spaces]

    def on_step(self, step, obs, action, reward, done, value=None):
        """
        Used as a callback by extending agents.
        Note that here "step" refers to update step, rather than agent timestep

        NB! Agent will overwrite previous batch without warning
        Keeping track of memory state is up to extending subclasses
        """
        step = step % self.traj_len
        self.batch_ptr = self.batch_ptr % self.batch_sz
        bs, be = self.batch_ptr, self.batch_ptr + reward.shape[0]

        self.dones[step, bs:be] = done
        self.rewards[step, bs:be] = reward

        if value is not None:
            self.values[step, bs:be] = value

        for i in range(len(obs)):
            self.obs[i][step, bs:be] = obs[i]

        for i in range(len(action)):
            self.acts[i][step, bs:be] = action[i]

        if (step+1) % self.traj_len == 0:
            # finished one trajectory
            for i in range(len(obs)):
                self.last_obs[i][bs:be] = self.next_obs[i]
            self.batch_ptr += reward.shape[0]

        if self.batch_ready():
            self.n_batches += 1

    def batch_ready(self):
        """
        Returns true if on_step was called batch_sz times
        """
        return self.batch_ptr == self.batch_sz



import gin
import tensorflow.compat.v1 as tf
from reaver.models.base import MultiPolicy


@gin.configurable
class SC2MultiPolicy(MultiPolicy):
    def __init__(self, act_spec, logits):
        super().__init__(act_spec, logits)

        args_mask = tf.constant(act_spec.spaces[0].args_mask, dtype=tf.float32)
        act_args_mask = tf.gather(args_mask, self.inputs[0])
        act_args_mask = tf.transpose(act_args_mask, [1, 0])

        self.logli = self.dists[0].log_prob(self.inputs[0])
        for i in range(1, len(self.dists)):
            self.logli += act_args_mask[i-1] * self.dists[i].log_prob(self.inputs[i])



import os
import math
import logging
import argparse
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from itertools import zip_longest
from tensorboard.backend.event_processing.event_accumulator import EventAccumulator


def plot_from_summaries(summaries_path, title=None, samples_per_update=512, updates_per_log=100):
    acc = EventAccumulator(summaries_path)
    acc.Reload()

    rews_mean = np.array([s[2] for s in acc.Scalars('Rewards/Mean')])
    rews_std = np.array([s[2] for s in acc.Scalars('Rewards/Std')])
    x = samples_per_update * updates_per_log * np.arange(0, len(rews_mean))

    if not title:
        title = summaries_path.split('/')[-1].split('_')[0]

    plt.plot(x, rews_mean)
    plt.fill_between(x, rews_mean - rews_std, rews_mean + rews_std, alpha=0.2)
    plt.xlabel('Samples')
    plt.ylabel('Episode Rewards')
    plt.title(title)
    plt.xlim([0, x[-1]+1])
    plt.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('logdir', type=str)
    parser.add_argument('--titles', nargs='*', default=[])
    parser.add_argument('--samples_per_update', type=int, default=512)
    parser.add_argument('--updates_per_log', type=int, default=100)
    args = parser.parse_args()

    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
    logging.getLogger('tensorflow').setLevel(logging.ERROR)

    plt.style.use('seaborn')
    mpl.rcParams['figure.figsize'] = (10, 10)

    paths = [os.path.join(args.logdir, p) for p in os.listdir(args.logdir)]

    assert len(paths) >= len(args.titles), "More titles than summaries"

    n_plots = len(paths)
    n_rows = math.ceil(n_plots / 2)
    for idx, (path, title) in enumerate(zip_longest(paths, args.titles)):
        if n_plots > 1:
            plt.subplot(n_rows, 2, 1 + idx)
        plot_from_summaries(path, title, args.samples_per_update, args.updates_per_log)
    plt.tight_layout()
    plt.show()



import numpy as np
from typing import List


class Space:
    """
    Holds information about any generic space
    In essence is a simplification of gym.spaces module into a single endpoint
    """
    def __init__(self, shape=(), dtype=np.int32, domain=(0, 1), categorical=False, name=None):
        self.name = name
        self.shape, self.dtype = shape, dtype
        self.categorical, (self.lo, self.hi) = categorical, domain

    def is_discrete(self) -> bool:
        """
        Space is considered continuous if its values are only ints
        """
        return np.issubdtype(self.dtype, np.integer)

    def is_continuous(self) -> bool:
        """
        Space is considered continuous if its values can be floats
        """
        return np.issubdtype(self.dtype, np.floating)

    def is_spatial(self) -> bool:
        """
        Space is considered spacial if it has three-dimensional shape HxWxC
        """
        return len(self.shape) > 1 or type(self.hi) in [list, tuple]

    def size(self) -> int:
        """
        Number of labels if categorical
        Number of intervals if discrete (can have multiple in one space)
        Number of mean and log std.dev if continuous

        Meant to be used to determine size of logit outputs in models
        """
        if self.is_discrete() and self.categorical:
            if self.is_spatial():
                return self.hi
            return self.hi - self.lo

        sz = 1
        if len(self.shape) == 1:
            sz = self.shape[0]

        return sz

    def sample(self, n=1):
        """
        Sample from this space. Useful for random agent, for example.
        """
        if self.is_discrete():
            return np.random.randint(self.lo, self.hi+1, (n, ) + self.shape)

        if self.is_continuous():
            return np.random.uniform(self.lo, self.hi+1e-10, (n, ) + self.shape)

    def __repr__(self):
        mid = str(self.shape)
        if self.categorical:
            mid += ", cat: " + str(self.hi)
        return "Space(%s, %s, %s)" % (self.name, mid, str(self.dtype).strip("<class>' "))


class Spec:
    """
    Convenience class to hold a list of spaces, can be used as an iterable
    A typical environment is expected to have one observation spec and one action spec

    Note: Every spec is expected to have a list of spaces, even if there is only one space
    """
    def __init__(self, spaces: List[Space], name=None):
        self.name, self.spaces = name, spaces
        for i, space in enumerate(self.spaces):
            if not space.name:
                space.name = str(i)

    def sample(self, n=1):
        return [space.sample(n) for space in self.spaces]

    def __repr__(self):
        return "Spec: %s\n%s" % (self.name, "\n".join(map(str, self.spaces)))

    def __iter__(self):
        return (space for space in self.spaces)

    def __len__(self):
        return len(self.spaces)



from abc import ABC, abstractmethod
from reaver.utils.typing import *
from .spec import Spec


class Env(ABC):
    """
    Abstract Base Class for all environments supported by Reaver
    Acts as a glue between the agents, models and envs modules
    Implementing class can be a simple wrapper (e.g. over openAI Gym)

    Note: observation / action specs contain a list of spaces,
          this is implicitly assumed across all Reaver components
    """
    def __init__(self, _id: str, render=False, reset_done=True, max_ep_len=None):
        self.id = _id
        self.render = render
        self.reset_done = reset_done
        self.max_ep_len = max_ep_len if max_ep_len else float('inf')

    @abstractmethod
    def start(self) -> None: ...

    @abstractmethod
    def step(self, action: Action) -> Tuple[Observation, Reward, Done]: ...

    @abstractmethod
    def reset(self) -> Observation: ...

    @abstractmethod
    def stop(self) -> None: ...

    @abstractmethod
    def obs_spec(self) -> Spec: ...

    @abstractmethod
    def act_spec(self) -> Spec: ...



from .base import *
from .random import RandomAgent
from .a2c import AdvantageActorCriticAgent
from .ppo import ProximalPolicyOptimizationAgent

A2C = AdvantageActorCriticAgent
PPO = ProximalPolicyOptimizationAgent

registry = {
    'a2c': A2C,
    'ppo': PPO
}



import gin
import tensorflow.compat.v1 as tf


@gin.configurable
class MultiPolicy:
    def __init__(self, act_spec, logits):
        self.logits = logits
        self.inputs = [tf.placeholder(s.dtype, [None, *s.shape]) for s in act_spec]

        self.dists = [self.make_dist(s, l) for s, l in zip(act_spec.spaces, logits)]

        self.entropy = sum([dist.entropy() for dist in self.dists])
        self.logli = sum([dist.log_prob(act) for dist, act in zip(self.dists, self.inputs)])

        self.sample = [dist.sample() for dist in self.dists]

    @staticmethod
    def make_dist(space, logits):
        # tfp is really heavy on init, better to lazy load
        import tensorflow_probability as tfp

        if space.is_continuous():
            mu, logstd = tf.split(logits, 2, axis=-1)
            return tfp.distributions.MultivariateNormalDiag(mu, tf.exp(logstd))
        else:
            return tfp.distributions.Categorical(logits)



import os

SC2_MINIGAMES_ALIASES = {
    'beacon': 'MoveToBeacon',
    'shards': 'CollectMineralShards',
    'roaches': 'DefeatRoaches',
    'blings': 'DefeatZerglingsAndBanelings',
    'lings': 'FindAndDefeatZerglings',
    'minerals': 'CollectMineralsAndGas',
    'marines': 'BuildMarines',
}

GYM_CONTINUOUS = [
    'LunarLanderContinuous-v2',
    'BipedalWalker-v2',
    'CarRacing-v0',
    'MountainCarContinuous-v0',
    'Pendulum-v0',
    'Acrobot-v1'
]

MUJOCO_ENVS = [
    'Reacher-v2',
    'Pusher-v2',
    'Thrower-v2',
    'Striker-v2',
    'InvertedPendulum-v2',
    'InvertedDoublePendulum-v2',
    'HalfCheetah-v2',
    'Hopper-v2',
    'Swimmer-v2',
    'Walker2d-v2',
    'Ant-v2',
    'Humanoid-v2',
    'HumanoidStandup-v2'
]

ATARI_GAMES = list(map(lambda name: ''.join([g.capitalize() for g in name.split('_')]), [
    'air_raid', 'alien', 'amidar', 'assault', 'asterix', 'asteroids', 'atlantis',
    'bank_heist', 'battle_zone', 'beam_rider', 'berzerk', 'bowling', 'boxing', 'breakout',
    'carnival', 'centipede', 'chopper_command', 'crazy_climber', 'demon_attack', 'double_dunk',
    'elevator_action', 'enduro', 'fishing_derby', 'freeway', 'frostbite', 'gopher', 'gravitar',
    'hero', 'ice_hockey', 'jamesbond', 'journey_escape', 'kangaroo', 'krull', 'kung_fu_master',
    'montezuma_revenge', 'ms_pacman', 'name_this_game', 'phoenix', 'pitfall', 'pong', 'pooyan',
    'solaris', 'space_invaders', 'star_gunner', 'tennis', 'time_pilot', 'tutankham', 'up_n_down',
    'private_eye', 'qbert', 'riverraid', 'road_runner', 'robotank', 'seaquest', 'skiing',
    'venture', 'video_pinball', 'wizard_of_wor', 'yars_revenge', 'zaxxon']))

ATARI_ENVS = ['%s-v0' % name for name in ATARI_GAMES]
ATARI_ENVS += ['%s-v4' % name for name in ATARI_GAMES]
ATARI_ENVS += ['%sDeterministic-v0' % name for name in ATARI_GAMES]
ATARI_ENVS += ['%sDeterministic-v4' % name for name in ATARI_GAMES]
ATARI_ENVS += ['%sNoFrameskip-v0' % name for name in ATARI_GAMES]
ATARI_ENVS += ['%sNoFrameskip-v4' % name for name in ATARI_GAMES]


def find_configs(env_name, base_path=''):
    if '-v' not in env_name:
        return filter_exists(['sc2/base.gin', 'sc2/%s.gin' % env_name], base_path)

    if env_name in GYM_CONTINUOUS:
        return filter_exists(['gym/base.gin', 'gym/continuous.gin', 'gym/%s.gin' % env_name], base_path)

    if env_name in MUJOCO_ENVS:
        return filter_exists(['mujoco/base.gin', 'mujoco/%s.gin' % env_name], base_path)

    if env_name in ATARI_ENVS:
        return filter_exists(['atari/base.gin', 'atari/%s.gin' % env_name], base_path)

    return filter_exists(['gym/base.gin', 'gym/%s.gin' % env_name], base_path)


def filter_exists(filenames, base_path):
    full_paths = [os.path.join(base_path, 'configs', fl) for fl in filenames]
    full_paths = [fl for fl in full_paths if os.path.exists(fl)]
    return full_paths



import os
import sys
import time
import numpy as np
from collections import deque, namedtuple


class Logger:
    def on_start(self): ...

    def on_step(self, step, rewards, dones): ...

    def on_update(self, step, loss_terms, grads_norm, returns, adv, next_value): ...

    def on_finish(self): ...


class StreamLogger(Logger):
    def __init__(self, n_envs, log_freq=100, rew_avg_eps=100, sess_mgr=None, log_file_path=None):
        self.n_envs = n_envs
        self.log_freq = log_freq
        self.rew_avg_eps = rew_avg_eps

        self.env_eps = [0]*n_envs
        self.env_rews = [0]*n_envs
        self.ep_rews_sum = deque([], maxlen=self.rew_avg_eps)

        self.run_time = 0
        self.start_time = None

        self.sess_mgr = sess_mgr
        self.streams = [sys.stdout]
        self.log_file_path = None
        if self.sess_mgr.training_enabled:
            self.log_file_path = log_file_path

        ColumnParams = namedtuple("ColumnParams", ["abbr", "width", "precision"])
        self.col_params = dict(
            runtime=ColumnParams("T", 6, 0),
            frames=ColumnParams("Fr", 9, 0),
            episodes=ColumnParams("Ep", 6, 0),
            updates=ColumnParams("Up", 6, 0),
            ep_rews_mean=ColumnParams("RMe", 7, 2),
            ep_rews_std=ColumnParams("RSd", 7, 2),
            ep_rews_max=ColumnParams("RMa", 7, 2),
            ep_rews_min=ColumnParams("RMi", 7, 2),
            policy_loss=ColumnParams("Pl", 8, 3),
            value_loss=ColumnParams("Vl", 8, 3),
            entropy_loss=ColumnParams("El", 6, 4),
            grads_norm=ColumnParams("Gr", 8, 3),
            frames_per_second=ColumnParams("Fps", 5, 0),
        )

        self.col_fmt = "| {abbr} {value:{width}.{precision}f} "

    def on_step(self, step, rewards, dones):
        self.env_rews += rewards
        for i in range(self.n_envs):
            if not dones[i]:
                continue
            self.ep_rews_sum.append(self.env_rews[i])
            self.env_rews[i] = 0
            self.env_eps[i] += 1

    def on_update(self, step, loss_terms, grads_norm, returns, adv, next_value):
        if step > 1 and step % self.log_freq:
            return

        frames = step * np.prod(returns.shape)
        run_time = max(1, int(time.time() - self.start_time)) + self.run_time
        ep_rews = np.array(self.ep_rews_sum or [0])

        logs = dict(
            runtime=run_time,
            frames=frames,
            updates=step,
            episodes=int(np.sum(self.env_eps)),
            frames_per_second=frames // run_time,
            ep_rews_mean=ep_rews.mean(),
            ep_rews_std=ep_rews.std(),
            ep_rews_max=ep_rews.max(),
            ep_rews_min=ep_rews.min(),
            policy_loss=loss_terms[0],
            value_loss=loss_terms[1],
            entropy_loss=loss_terms[2],
            grads_norm=grads_norm,
        )

        self.stream_logs(logs)
        if self.sess_mgr:
            self.summarize_logs(logs)

    def stream_logs(self, logs):
        log_str = ""
        for key, params in self.col_params.items():
            abbr, width, precision = params.abbr, params.width, params.precision
            log_str += self.col_fmt.format(abbr=abbr, value=logs[key], width=width, precision=precision)
        log_str += "|"

        for stream in self.streams:
            print(log_str, file=stream)
            stream.flush()

    def summarize_logs(self, logs):
        losses = [logs['policy_loss'], logs['value_loss'], logs['entropy_loss']]
        rews = [logs['ep_rews_mean'], logs['ep_rews_std'], logs['ep_rews_max'], logs['ep_rews_min']]

        self.sess_mgr.add_summaries(['Mean', 'Std', 'Max', 'Min'], rews, 'Rewards', logs['updates'])
        self.sess_mgr.add_summaries(['Policy', 'Value', 'Entropy'], losses, 'Losses', logs['updates'])
        self.sess_mgr.add_summary('Grads Norm', logs['grads_norm'], 'Losses', logs['updates'])

    def on_start(self):
        self.start_time = time.time()
        if not self.log_file_path:
            return

        self.restore_logs()
        self.streams.append(open(self.log_file_path, 'a+'))

    def on_finish(self):
        if len(self.streams) > 1:
            self.streams[1].close()

    def restore_logs(self):
        if not os.path.isfile(self.log_file_path):
            return

        with open(self.log_file_path, 'r') as fl:
            last_line = fl.readlines()[-1]
        logs = last_line.split(" | ")
        self.run_time = int(logs[0].split(" ")[-1])
        self.env_eps.append(int(logs[2].split(" ")[-1]))


class AgentDebugLogger(Logger):
    def __init__(self, agent, log_freq=100, debug_steps=10):
        self.agent = agent
        self.log_freq = log_freq
        self.debug_steps = debug_steps

    def on_update(self, step, loss_terms, grads_norm, returns, adv, next_value):
        update_step = (step + 1) // self.agent.traj_len
        if update_step > 1 and update_step % self.log_freq:
            return

        np.set_printoptions(suppress=True, precision=2)
        n_steps = min(self.debug_steps, self.agent.traj_len)

        print()
        print("First Env For Last %d Steps:" % n_steps)
        print("Dones      ", self.agent.dones[-n_steps:, 0].flatten().astype(int))
        print("Rewards    ", self.agent.rewards[-n_steps:, 0].flatten())
        print("Values     ", self.agent.values[-n_steps:, 0].flatten(), round(next_value[0], 3))
        print("Returns    ", returns[-n_steps:, 0].flatten())
        print("Advs       ", adv[-n_steps:, 0].flatten())

        sys.stdout.flush()



import gin.tf
import numpy as np
import tensorflow.compat.v1 as tf
from abc import abstractmethod

from reaver.envs.base import Spec
from reaver.agents.base import MemoryAgent
from reaver.utils import Logger
from reaver.utils.tensorflow import SessionManager
from reaver.utils.typing import ModelBuilder, PolicyType

DEFAULTS = dict(
    model_fn=None,
    policy_cls=None,
    optimizer=None,
    learning_rate=0.0003,
    value_coef=0.5,
    entropy_coef=0.01,
    traj_len=16,
    batch_sz=16,
    discount=0.99,
    gae_lambda=0.95,
    clip_rewards=0.0,
    clip_grads_norm=0.0,
    normalize_returns=False,
    normalize_advantages=False,
)


@gin.configurable('ACAgent')
class ActorCriticAgent(MemoryAgent):
    """
    Abstract class, unifies deep actor critic functionality
    Handles on_step callbacks, either updating current batch
    or executing one training step if the batch is ready

    Extending classes only need to implement loss_fn method
    """
    def __init__(
        self,
        obs_spec: Spec,
        act_spec: Spec,
        model_fn: ModelBuilder=None,
        policy_cls: PolicyType=None,
        sess_mgr: SessionManager=None,
        optimizer: tf.train.Optimizer=None,
        value_coef=DEFAULTS['value_coef'],
        entropy_coef=DEFAULTS['entropy_coef'],
        traj_len=DEFAULTS['traj_len'],
        batch_sz=DEFAULTS['batch_sz'],
        discount=DEFAULTS['discount'],
        gae_lambda=DEFAULTS['gae_lambda'],
        clip_rewards=DEFAULTS['clip_rewards'],
        clip_grads_norm=DEFAULTS['clip_grads_norm'],
        normalize_returns=DEFAULTS['normalize_returns'],
        normalize_advantages=DEFAULTS['normalize_advantages'],
    ):
        MemoryAgent.__init__(self, obs_spec, act_spec, traj_len, batch_sz)

        if not sess_mgr:
            sess_mgr = SessionManager()

        if not optimizer:
            optimizer = tf.train.AdamOptimizer(learning_rate=DEFAULTS['learning_rate'])

        self.sess_mgr = sess_mgr
        self.value_coef = value_coef
        self.entropy_coef = entropy_coef
        self.discount = discount
        self.gae_lambda = gae_lambda
        self.clip_rewards = clip_rewards
        self.normalize_returns = normalize_returns
        self.normalize_advantages = normalize_advantages

        self.model = model_fn(obs_spec, act_spec)
        self.value = self.model.outputs[-1]
        self.policy = policy_cls(act_spec, self.model.outputs[:-1])
        self.loss_op, self.loss_terms, self.loss_inputs = self.loss_fn()

        grads, vars = zip(*optimizer.compute_gradients(self.loss_op))
        self.grads_norm = tf.global_norm(grads)
        if clip_grads_norm > 0.:
            grads, _ = tf.clip_by_global_norm(grads, clip_grads_norm, self.grads_norm)
        self.train_op = optimizer.apply_gradients(zip(grads, vars), global_step=sess_mgr.global_step)
        self.minimize_ops = self.make_minimize_ops()

        sess_mgr.restore_or_init()
        self.n_batches = sess_mgr.start_step
        self.start_step = sess_mgr.start_step * traj_len

        self.logger = Logger()

    def get_action_and_value(self, obs):
        return self.sess_mgr.run([self.policy.sample, self.value], self.model.inputs, obs)

    def get_action(self, obs):
        return self.sess_mgr.run(self.policy.sample, self.model.inputs, obs)

    def on_step(self, step, obs, action, reward, done, value=None):
        MemoryAgent.on_step(self, step, obs, action, reward, done, value)
        self.logger.on_step(step, reward, done)

        if not self.batch_ready():
            return

        next_values = self.sess_mgr.run(self.value, self.model.inputs, self.last_obs)
        adv, returns = self.compute_advantages_and_returns(next_values)

        loss_terms, grads_norm = self.minimize(adv, returns)

        self.sess_mgr.on_update(self.n_batches)
        self.logger.on_update(self.n_batches, loss_terms, grads_norm, returns, adv, next_values)

    def minimize(self, advantages, returns):
        inputs = self.obs + self.acts + [advantages, returns]
        inputs = [a.reshape(-1, *a.shape[2:]) for a in inputs]
        tf_inputs = self.model.inputs + self.policy.inputs + self.loss_inputs

        loss_terms, grads_norm, *_ = self.sess_mgr.run(self.minimize_ops, tf_inputs, inputs)

        return loss_terms, grads_norm

    def compute_advantages_and_returns(self, bootstrap_value):
        """
        GAE can help with reducing variance of policy gradient estimates
        """
        if self.clip_rewards > 0.0:
            np.clip(self.rewards, -self.clip_rewards, self.clip_rewards, out=self.rewards)

        rewards = self.rewards.copy()
        rewards[-1] += (1-self.dones[-1]) * self.discount * bootstrap_value

        masked_discounts = self.discount * (1-self.dones)

        returns = self.discounted_cumsum(rewards, masked_discounts)

        if self.gae_lambda > 0.:
            values = np.append(self.values, np.expand_dims(bootstrap_value, 0), axis=0)
            # d_t = r_t + g * V(s_{t+1}) - V(s_t)
            deltas = self.rewards + masked_discounts * values[1:] - values[:-1]
            adv = self.discounted_cumsum(deltas, self.gae_lambda * masked_discounts)
        else:
            adv = returns - self.values

        if self.normalize_advantages:
            adv = (adv - adv.mean()) / (adv.std() + 1e-10)

        if self.normalize_returns:
            returns = (returns - returns.mean()) / (returns.std() + 1e-10)

        return adv, returns

    def on_start(self):
        self.logger.on_start()

    def on_finish(self):
        self.logger.on_finish()

    def make_minimize_ops(self):
        ops = [self.loss_terms, self.grads_norm]
        if self.sess_mgr.training_enabled:
            ops.append(self.train_op)
        # appending extra model update ops (e.g. running stats)
        # note: this will most likely break if model.compile() is used
        ops.extend(self.model.get_updates_for(None))
        return ops

    @staticmethod
    def discounted_cumsum(x, discount):
        y = np.zeros_like(x)
        y[-1] = x[-1]
        for t in range(x.shape[0]-2, -1, -1):
            y[t] = x[t] + discount[t] * y[t+1]
        return y

    @abstractmethod
    def loss_fn(self): ...



import sys
import gin
import numpy as np
from absl import flags
from pysc2.lib import actions
from pysc2.lib import features
from pysc2.lib import protocol
from pysc2.env.environment import StepType
from . import Env, Spec, Space

ACTIONS_MINIGAMES, ACTIONS_MINIGAMES_ALL, ACTIONS_ALL = ['minigames', 'minigames_all', 'all']


@gin.configurable
class SC2Env(Env):
    """
    'minigames' action set is enough to solve all minigames listed in SC2LE
    'minigames_all' expands that set with actions that may improve end results, but will drop performance
    'all' is the full action set, only necessary for generic agent playing full game with all three races

    You can also specify your own action set in the gin config file under SC2Env.action_ids
    Full list of available actions https://github.com/deepmind/pysc2/blob/master/pysc2/lib/actions.py#L447-L1008
    """
    def __init__(
        self,
        map_name='MoveToBeacon',
        render=False,
        reset_done=True,
        max_ep_len=None,
        spatial_dim=16,
        step_mul=8,
        obs_features=None,
        action_ids=ACTIONS_MINIGAMES
    ):
        super().__init__(map_name, render, reset_done, max_ep_len)

        self.step_mul = step_mul
        self.spatial_dim = spatial_dim
        self._env = None

        # sensible action set for all minigames
        if not action_ids or action_ids in [ACTIONS_MINIGAMES, ACTIONS_MINIGAMES_ALL]:
            action_ids = [0, 1, 2, 3, 4, 6, 7, 12, 13, 42, 44, 50, 91, 183, 234, 309, 331, 332, 333, 334, 451, 452, 490]

        # some additional actions for minigames (not necessary to solve)
        if action_ids == ACTIONS_MINIGAMES_ALL:
            action_ids += [11, 71, 72, 73, 74, 79, 140, 168, 239, 261, 264, 269, 274, 318, 335, 336, 453, 477]

        # full action space, including outdated / unusable to current race / usable only in certain cases
        if action_ids == ACTIONS_ALL:
            action_ids = [f.id for f in actions.FUNCTIONS]

        # by default use majority of obs features, except for some that are unnecessary for minigames
        # e.g. race-specific like creep and shields or redundant like player_id
        if not obs_features:
            obs_features = {
                'screen': ['player_relative', 'selected', 'visibility_map', 'unit_hit_points_ratio', 'unit_density'],
                'minimap': ['player_relative', 'selected', 'visibility_map', 'camera'],
                # available actions should always be present and in first position
                'non-spatial': ['available_actions', 'player']}

        self.act_wrapper = ActionWrapper(spatial_dim, action_ids)
        self.obs_wrapper = ObservationWrapper(obs_features, action_ids)

    def start(self):
        # importing here to lazy-load
        from pysc2.env import sc2_env

        # fail-safe if executed not as absl app
        if not flags.FLAGS.is_parsed():
            flags.FLAGS(sys.argv)

        self._env = sc2_env.SC2Env(
            map_name=self.id,
            visualize=self.render,
            agent_interface_format=[features.parse_agent_interface_format(
                feature_screen=self.spatial_dim,
                feature_minimap=self.spatial_dim,
                rgb_screen=None,
                rgb_minimap=None
            )],
            step_mul=self.step_mul,
            players=[sc2_env.Agent(sc2_env.Race.terran)])

    def step(self, action):
        try:
            obs, reward, done = self.obs_wrapper(self._env.step(self.act_wrapper(action)))
        except protocol.ConnectionError:
            # hacky fix from websocket timeout issue...
            # this results in faulty reward signals, but I guess it beats completely crashing...
            self.restart()
            return self.reset(), 0, 1

        if done and self.reset_done:
            obs = self.reset()

        return obs, reward, done

    def reset(self):
        try:
            obs, reward, done = self.obs_wrapper(self._env.reset())
        except protocol.ConnectionError:
            # hacky fix from websocket timeout issue...
            # this results in faulty reward signals, but I guess it beats completely crashing...
            self.restart()
            return self.reset()

        return obs

    def stop(self):
        self._env.close()

    def restart(self):
        self.stop()
        self.start()

    def obs_spec(self):
        if not self.obs_wrapper.spec:
            self.make_specs()
        return self.obs_wrapper.spec

    def act_spec(self):
        if not self.act_wrapper.spec:
            self.make_specs()
        return self.act_wrapper.spec

    def make_specs(self):
        # importing here to lazy-load
        from pysc2.env import mock_sc2_env
        mock_env = mock_sc2_env.SC2TestEnv(map_name=self.id, agent_interface_format=[
            features.parse_agent_interface_format(feature_screen=self.spatial_dim, feature_minimap=self.spatial_dim)])
        self.act_wrapper.make_spec(mock_env.action_spec())
        self.obs_wrapper.make_spec(mock_env.observation_spec())
        mock_env.close()


class ObservationWrapper:
    def __init__(self, _features=None, action_ids=None):
        self.spec = None
        self.features = _features
        self.action_ids = action_ids

        screen_feature_to_idx = {feat: idx for idx, feat in enumerate(features.SCREEN_FEATURES._fields)}
        minimap_feature_to_idx = {feat: idx for idx, feat in enumerate(features.MINIMAP_FEATURES._fields)}

        self.feature_masks = {
            'screen': [screen_feature_to_idx[f] for f in _features['screen']],
            'minimap': [minimap_feature_to_idx[f] for f in _features['minimap']]
        }

    def __call__(self, timestep):
        ts = timestep[0]
        obs, reward, done = ts.observation, ts.reward, ts.step_type == StepType.LAST

        obs_wrapped = [
            obs['feature_screen'][self.feature_masks['screen']],
            obs['feature_minimap'][self.feature_masks['minimap']]
        ]
        for feat_name in self.features['non-spatial']:
            if feat_name == 'available_actions':
                fn_ids_idxs = [i for i, fn_id in enumerate(self.action_ids) if fn_id in obs[feat_name]]
                mask = np.zeros((len(self.action_ids),), dtype=np.int32)
                mask[fn_ids_idxs] = 1
                obs[feat_name] = mask
            obs_wrapped.append(obs[feat_name])

        return obs_wrapped, reward, done

    def make_spec(self, spec):
        spec = spec[0]

        default_dims = {
            'available_actions': (len(self.action_ids), ),
        }

        screen_shape = (len(self.features['screen']), *spec['feature_screen'][1:])
        minimap_shape = (len(self.features['minimap']), *spec['feature_minimap'][1:])
        screen_dims = get_spatial_dims(self.features['screen'], features.SCREEN_FEATURES)
        minimap_dims = get_spatial_dims(self.features['minimap'], features.MINIMAP_FEATURES)

        spaces = [
            SC2Space(screen_shape, 'screen', self.features['screen'], screen_dims),
            SC2Space(minimap_shape, 'minimap', self.features['minimap'], minimap_dims),
        ]

        for feat in self.features['non-spatial']:
            if 0 in spec[feat]:
                spec[feat] = default_dims[feat]
            spaces.append(Space(spec[feat], name=feat))

        self.spec = Spec(spaces, 'Observation')


class ActionWrapper:
    def __init__(self, spatial_dim, action_ids, args=None):
        self.spec = None
        if not args:
            args = [
                'screen',
                'minimap',
                'screen2',
                'queued',
                'control_group_act',
                'control_group_id',
                'select_add',
                'select_point_act',
                'select_unit_act',
                # 'select_unit_id'
                'select_worker',
                'build_queue_id',
                # 'unload_id'
            ]
        self.func_ids = action_ids
        self.args, self.spatial_dim = args, spatial_dim

    def __call__(self, action):
        defaults = {
            'control_group_act': 0,
            'control_group_id': 0,
            'select_point_act': 0,
            'select_unit_act': 0,
            'select_unit_id': 0,
            'build_queue_id': 0,
            'unload_id': 0,
        }
        fn_id_idx, args = action.pop(0), []
        fn_id = self.func_ids[fn_id_idx]
        for arg_type in actions.FUNCTIONS[fn_id].args:
            arg_name = arg_type.name
            if arg_name in self.args:
                arg = action[self.args.index(arg_name)]
                # pysc2 expects all args in their separate lists
                if type(arg) not in [list, tuple]:
                    arg = [arg]
                # pysc2 expects spatial coords, but we have flattened => attempt to fix
                if len(arg_type.sizes) > 1 and len(arg) == 1:
                    arg = [arg[0] % self.spatial_dim, arg[0] // self.spatial_dim]
                args.append(arg)
            else:
                args.append([defaults[arg_name]])

        return [actions.FunctionCall(fn_id, args)]

    def make_spec(self, spec):
        spec = spec[0]

        spaces = [SC2FuncIdSpace(self.func_ids, self.args)]
        for arg_name in self.args:
            arg = getattr(spec.types, arg_name)
            if len(arg.sizes) > 1:
                spaces.append(Space(domain=(0, arg.sizes), categorical=True, name=arg_name))
            else:
                spaces.append(Space(domain=(0, arg.sizes[0]), categorical=True, name=arg_name))

        self.spec = Spec(spaces, "Action")


class SC2Space(Space):
    def __init__(self, shape, name, spatial_feats=None, spatial_dims=None):
        if spatial_feats:
            name += "{%s}" % ", ".join(spatial_feats)
        self.spatial_feats, self.spatial_dims = spatial_feats, spatial_dims

        super().__init__(shape, name=name)


class SC2FuncIdSpace(Space):
    def __init__(self, func_ids, args):
        super().__init__(domain=(0, len(func_ids)), categorical=True, name="function_id")
        self.args_mask = []
        for fn_id in func_ids:
            fn_id_args = [arg_type.name for arg_type in actions.FUNCTIONS[fn_id].args]
            self.args_mask.append([arg in fn_id_args for arg in args])


def get_spatial_dims(feat_names, feats):
    feats_dims = []
    for feat_name in feat_names:
        feat = getattr(feats, feat_name)
        feats_dims.append(1)
        if feat.type == features.FeatureType.CATEGORICAL:
            feats_dims[-1] = feat.scale
    return feats_dims



import setuptools

with open("README.md", 'r') as f:
    long_description = f.read()

setuptools.setup(
    name='reaver',
    version='2.1.9',
    author='Roman Ring',
    author_email='inoryy@gmail.com',
    description='Reaver: Modular Deep Reinforcement Learning Framework. Focused on StarCraft II. '
                'Supports Gym, Atari, and MuJoCo. Matches reference results.',
    keywords='reaver starcraft2 gym atari mujoco tensorflow reinforcement learning neural network',
    include_package_data=True,
    packages=setuptools.find_packages(),
    install_requires=[
        'PySC2 >= 3.0.0',
        'gin-config >= 0.3.0',
        'tensorflow >= 2.1.0',
        'tensorflow-probability >= 0.9.0'
    ],
    extras_require={
        'gym': [
            'PyOpenGL',
            'gym >= 0.9',
            'opencv-python'
        ],
        'atari': [
            'Pillow',
            'PyOpenGL',
            'gym >= 0.9',
            'atari_py >= 0.1.4',
        ],
        'mujoco': [
            'imageio',
            'gym >= 0.9',
            'mujoco_py >= 1.50',
        ]
    },
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    url='https://github.com/inoryy/reaver',
)



# ########################################################################################## #
# Copied from https://github.com/google/dopamine/blob/master/dopamine/atari/preprocessing.py #
# Changes (minimal): added proxy close method, reformatted, removed some unused pieces       #
# ########################################################################################## #

# coding=utf-8
# Copyright 2018 The Dopamine Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A class implementing minimal Atari 2600 preprocessing.

This includes:
  . Emitting a terminal signal when losing a life (optional).
  . Frame skipping and color pooling.
  . Resizing the image before it is provided to the agent.
"""

import cv2
import gin.tf
import numpy as np
from gym.spaces.box import Box


@gin.configurable
class AtariPreprocessing(object):
    """A class implementing image preprocessing for Atari 2600 agents.

    Specifically, this provides the following subset from the JAIR paper
    (Bellemare et al., 2013) and Nature DQN paper (Mnih et al., 2015):

      * Frame skipping (defaults to 4).
      * Terminal signal when a life is lost (off by default).
      * Grayscale and max-pooling of the last two frames.
      * Downsample the screen to a square image (defaults to 84x84).

    More generally, this class follows the preprocessing guidelines set down in
    Machado et al. (2018), "Revisiting the Arcade Learning Environment:
    Evaluation Protocols and Open Problems for General Agents".
    """

    def __init__(self, environment, frame_skip=4, terminal_on_life_loss=False,
                 screen_size=84):
        """Constructor for an Atari 2600 preprocessor.

        Args:
          environment: Gym environment whose observations are preprocessed.
          frame_skip: int, the frequency at which the agent experiences the game.
          terminal_on_life_loss: bool, If True, the step() method returns
            is_terminal=True whenever a life is lost. See Mnih et al. 2015.
          screen_size: int, size of a resized Atari 2600 frame.

        Raises:
          ValueError: if frame_skip or screen_size are not strictly positive.
        """
        if frame_skip <= 0:
            raise ValueError('Frame skip should be strictly positive, got {}'.
                             format(frame_skip))
        if screen_size <= 0:
            raise ValueError('Target screen size should be strictly positive, got {}'.
                             format(screen_size))

        self.environment = environment
        self.terminal_on_life_loss = terminal_on_life_loss
        self.frame_skip = frame_skip
        self.screen_size = screen_size

        obs_dims = self.environment.observation_space
        # Stores temporary observations used for pooling over two successive
        # frames.
        self.screen_buffer = [
            np.empty((obs_dims.shape[0], obs_dims.shape[1]), dtype=np.uint8),
            np.empty((obs_dims.shape[0], obs_dims.shape[1]), dtype=np.uint8)
        ]

        self.game_over = False
        self.lives = 0  # Will need to be set by reset().

    @property
    def observation_space(self):
        # Return the observation space adjusted to match the shape of the processed
        # observations.
        return Box(low=0, high=255, shape=(self.screen_size, self.screen_size, 1),
                   dtype=np.uint8)

    @property
    def action_space(self):
        return self.environment.action_space

    @property
    def reward_range(self):
        return self.environment.reward_range

    @property
    def metadata(self):
        return self.environment.metadata

    def reset(self):
        """Resets the environment.

        Returns:
          observation: numpy array, the initial observation emitted by the
            environment.
        """
        self.environment.reset()
        self.lives = self.environment.ale.lives()
        self._fetch_grayscale_observation(self.screen_buffer[0])
        self.screen_buffer[1].fill(0)
        return self._pool_and_resize()

    def render(self, mode):
        """Renders the current screen, before preprocessing.

        This calls the Gym API's render() method.

        Args:
          mode: Mode argument for the environment's render() method.
            Valid values (str) are:
              'rgb_array': returns the raw ALE image.
              'human': renders to display via the Gym renderer.

        Returns:
          if mode='rgb_array': numpy array, the most recent screen.
          if mode='human': bool, whether the rendering was successful.
        """
        return self.environment.render(mode)

    def step(self, action):
        """Applies the given action in the environment.

        Remarks:

          * If a terminal state (from life loss or episode end) is reached, this may
            execute fewer than self.frame_skip steps in the environment.
          * Furthermore, in this case the returned observation may not contain valid
            image data and should be ignored.

        Args:
          action: The action to be executed.

        Returns:
          observation: numpy array, the observation following the action.
          reward: float, the reward following the action.
          is_terminal: bool, whether the environment has reached a terminal state.
            This is true when a life is lost and terminal_on_life_loss, or when the
            episode is over.
          info: Gym API's info data structure.
        """
        accumulated_reward = 0.

        for time_step in range(self.frame_skip):
            # We bypass the Gym observation altogether and directly fetch the
            # grayscale image from the ALE. This is a little faster.
            _, reward, game_over, info = self.environment.step(action)
            accumulated_reward += reward

            if self.terminal_on_life_loss:
                new_lives = self.environment.ale.lives()
                is_terminal = game_over or new_lives < self.lives
                self.lives = new_lives
            else:
                is_terminal = game_over

            if is_terminal:
                break
            # We max-pool over the last two frames, in grayscale.
            elif time_step >= self.frame_skip - 2:
                t = time_step - (self.frame_skip - 2)
                self._fetch_grayscale_observation(self.screen_buffer[t])

        # Pool the last two observations.
        observation = self._pool_and_resize()

        self.game_over = game_over
        return observation, accumulated_reward, is_terminal, info

    def close(self):
        self.environment.close()

    def _fetch_grayscale_observation(self, output):
        """Returns the current observation in grayscale.

        The returned observation is stored in 'output'.

        Args:
          output: numpy array, screen buffer to hold the returned observation.

        Returns:
          observation: numpy array, the current observation in grayscale.
        """
        self.environment.ale.getScreenGrayscale(output)
        return output

    def _pool_and_resize(self):
        """Transforms two frames into a Nature DQN observation.

        For efficiency, the transformation is done in-place in self.screen_buffer.

        Returns:
          transformed_screen: numpy array, pooled, resized screen.
        """
        # Pool if there are enough screens to do so.
        if self.frame_skip > 1:
            np.maximum(self.screen_buffer[0], self.screen_buffer[1],
                       out=self.screen_buffer[0])

        transformed_image = cv2.resize(self.screen_buffer[0],
                                       (self.screen_size, self.screen_size),
                                       interpolation=cv2.INTER_AREA)
        int_image = np.asarray(transformed_image, dtype=np.uint8)
        return np.expand_dims(int_image, axis=2)



import tensorflow.compat.v1 as tf
from tensorflow.compat.v1.keras.layers import Lambda, Layer


class RunningStatsNorm(Layer):
    """
    Normalizes inputs by running mean / std.dev statistics
    """
    def __init__(self, and_shift=True, and_scale=False, **kwargs):
        self.and_shift, self.and_scale = and_shift, and_scale
        self._ct = self._mu = self._var = None
        super().__init__(**kwargs)

    def build(self, input_shape):
        shape = (*input_shape[1:],)
        self._ct = self.add_weight('running_ct', (), initializer='zeros', trainable=False)
        self._mu = self.add_weight('running_mu', shape, initializer='zeros', trainable=False)
        self._var = self.add_weight('running_var', shape, initializer='ones', trainable=False)
        super().build(input_shape)

    def call(self, inputs, **kwargs):
        mu, var = self._update_stats(inputs)
        if self.and_shift:
            inputs -= mu
        if self.and_scale:
            inputs /= tf.sqrt(var)
        return inputs

    def _update_stats(self, x):
        ct = tf.maximum(1e-10, self._ct)

        ct_b = tf.to_float(tf.shape(x)[0])
        mu_b, var_b = tf.nn.moments(x, axes=[0])

        delta = mu_b - self._mu

        new_ct = ct + ct_b
        new_mu = self._mu + delta * ct_b / new_ct
        new_var = (self._var * ct + var_b * ct_b + delta ** 2 * ct * ct_b / new_ct) / new_ct

        self.add_update([
            tf.assign(self._ct, new_ct),
            tf.assign(self._mu, new_mu),
            tf.assign(self._var, new_var)
        ])

        return new_mu, new_var


class Variable(Layer):
    """
    Concatenate an extra trainable variable to the dense layer
    This variable is disconnected from the rest of the NN, including inputs
    """
    def __init__(self, **kwargs):
        self._var = None
        super().__init__(**kwargs)

    def build(self, input_shape):
        self._var = self.add_weight('var', (1, input_shape[-1]), initializer='zeros', trainable=True)
        super().build(input_shape)

    def call(self, inputs, **kwargs):
        """
        repeat _var N times to match inputs dim and then concatenate them
        """
        return tf.concat([inputs, tf.tile(self._var, (tf.shape(inputs)[0], 1))], axis=-1)


class Squeeze(Lambda):
    def __init__(self, axis=-1, name=None):
        Lambda.__init__(self, lambda x: tf.squeeze(x, axis=axis), name=name)


class Split(Lambda):
    def __init__(self, num_splits=2, axis=-1, name=None):
        Lambda.__init__(self, lambda x: tf.split(x, num_splits, axis=axis), name=name)


class Transpose(Lambda):
    def __init__(self, dims):
        Lambda.__init__(self, lambda x: tf.transpose(x, dims))


class Log(Lambda):
    def __init__(self):
        Lambda.__init__(self, lambda x: tf.log(x + 1e-10))


class Rescale(Lambda):
    def __init__(self, scale):
        Lambda.__init__(self, lambda x: tf.cast(x, tf.float32) * scale)


class Broadcast2D(Lambda):
    def __init__(self, size):
        Lambda.__init__(self, lambda x: tf.tile(tf.expand_dims(tf.expand_dims(x, 2), 3), [1, 1, size, size]))



import os
import gin
from datetime import datetime as dt


class Experiment:
    def __init__(self, results_dir, env_name, agent_name, name=None, restore=False):
        if not name:
            if restore:
                experiments = [e for e in os.listdir(results_dir) if env_name in e and agent_name in e]
                assert len(experiments) > 0, 'No experiment to restore'
                name = max(experiments, key=lambda p: os.path.getmtime(results_dir+'/'+p))
                name = '_'.join(name.split('_')[2:])
            else:
                name = dt.now().strftime("%y-%m-%d_%H-%M-%S")

        self.name = name
        self.restore = restore
        self.env_name = env_name
        self.agent_name = agent_name
        self.results_dir = results_dir

        os.makedirs(self.path, exist_ok=True)
        os.makedirs(self.path + '/summaries', exist_ok=True)
        os.makedirs(self.results_dir + '/summaries', exist_ok=True)
        if not os.path.exists(self.summaries_path):
            os.symlink('../%s/summaries' % self.full_name, self.summaries_path)

    @property
    def full_name(self):
        return '%s_%s_%s' % (self.env_name, self.agent_name, self.name)

    @property
    def path(self):
        return '%s/%s' % (self.results_dir, self.full_name)

    @property
    def config_path(self):
        return '%s/%s' % (self.path, 'config.gin')

    @property
    def log_path(self):
        return '%s/%s' % (self.path, 'train.log')

    @property
    def checkpoints_path(self):
        return self.path + '/checkpoints'

    @property
    def summaries_path(self):
        return '%s/summaries/%s' % (self.results_dir, self.full_name)

    def save_gin_config(self):
        with open(self.config_path, 'w') as cfg_file:
            cfg_file.write(gin.operative_config_str())

    def save_model_summary(self, model):
        with open(self.path + '/' + 'model_summary.txt', 'w') as fl:
            model.summary(print_fn=lambda line: print(line, file=fl))



import os
import gin
import tensorflow.compat.v1 as tf
from absl import app, flags

import reaver as rvr

flags.DEFINE_string('env', None, 'Either Gym env id or PySC2 map name to run agent in.')
flags.DEFINE_string('agent', 'a2c', 'Name of the agent. Must be one of (a2c, ppo).')

flags.DEFINE_bool('render', False, 'Whether to render first(!) env.')
flags.DEFINE_string('gpu', '0', 'GPU(s) id(s) to use. If not set TensorFlow will use CPU.')

flags.DEFINE_integer('n_envs', 4, 'Number of environments to execute in parallel.')
flags.DEFINE_integer('n_updates', 1000000, 'Number of train updates (1 update has batch_sz * traj_len samples).')

flags.DEFINE_integer('ckpt_freq', 500, 'Number of train updates per one checkpoint save.')
flags.DEFINE_integer('log_freq', 100, 'Number of train updates per one console log.')
flags.DEFINE_integer('log_eps_avg', 100, 'Number of episodes to average for performance stats.')
flags.DEFINE_integer('max_ep_len', None, 'Max number of steps an agent can take in an episode.')

flags.DEFINE_string('results_dir', 'results', 'Directory for model weights, train logs, etc.')
flags.DEFINE_string('experiment', None, 'Name of the experiment. Datetime by default.')

flags.DEFINE_multi_string('gin_files', [], 'List of path(s) to gin config(s).')
flags.DEFINE_multi_string('gin_bindings', [], 'Gin bindings to override config values.')

flags.DEFINE_bool('restore', False,
                  'Restore & continue previously executed experiment. '
                  'If experiment not specified then last modified is used.')
flags.DEFINE_bool('test', False,
                  'Run an agent in test mode: restore flag is set to true and number of envs set to 1'
                  'Loss is calculated, but gradients are not applied.'
                  'Checkpoints, summaries, log files are not updated, but console logger is enabled.')

flags.DEFINE_alias('e', 'env')
flags.DEFINE_alias('a', 'agent')
flags.DEFINE_alias('p', 'n_envs')
flags.DEFINE_alias('u', 'n_updates')
flags.DEFINE_alias('lf', 'log_freq')
flags.DEFINE_alias('cf', 'ckpt_freq')
flags.DEFINE_alias('la', 'log_eps_avg')
flags.DEFINE_alias('n', 'experiment')
flags.DEFINE_alias('g', 'gin_bindings')


def main(argv):
    tf.disable_eager_execution()
    tf.disable_v2_behavior()

    args = flags.FLAGS
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
    os.environ["CUDA_VISIBLE_DEVICES"] = args.gpu

    if args.env in rvr.utils.config.SC2_MINIGAMES_ALIASES:
        args.env = rvr.utils.config.SC2_MINIGAMES_ALIASES[args.env]

    if args.test:
        args.n_envs = 1
        args.log_freq = 1
        args.restore = True

    expt = rvr.utils.Experiment(args.results_dir, args.env, args.agent, args.experiment, args.restore)

    gin_files = rvr.utils.find_configs(args.env, os.path.dirname(os.path.abspath(__file__)))
    if args.restore:
        gin_files += [expt.config_path]
    gin_files += args.gin_files

    if not args.gpu:
        args.gin_bindings.append("build_cnn_nature.data_format = 'channels_last'")
        args.gin_bindings.append("build_fully_conv.data_format = 'channels_last'")

    gin.parse_config_files_and_bindings(gin_files, args.gin_bindings)
    args.n_envs = min(args.n_envs, gin.query_parameter('ACAgent.batch_sz'))

    sess = tf.Session(config=tf.ConfigProto(allow_soft_placement=True))
    sess_mgr = rvr.utils.tensorflow.SessionManager(sess, expt.path, args.ckpt_freq, training_enabled=not args.test)

    env_cls = rvr.envs.GymEnv if '-v' in args.env else rvr.envs.SC2Env
    env = env_cls(args.env, args.render, max_ep_len=args.max_ep_len)

    agent = rvr.agents.registry[args.agent](env.obs_spec(), env.act_spec(), sess_mgr=sess_mgr, n_envs=args.n_envs)
    agent.logger = rvr.utils.StreamLogger(args.n_envs, args.log_freq, args.log_eps_avg, sess_mgr, expt.log_path)

    if sess_mgr.training_enabled:
        expt.save_gin_config()
        expt.save_model_summary(agent.model)

    agent.run(env, args.n_updates * agent.traj_len * agent.batch_sz // args.n_envs)


if __name__ == '__main__':
    flags.mark_flag_as_required('env')
    app.run(main)
